"""Primary interface for analyzing the XROMM data using trained network"""

import os
import math

from itertools import combinations
import cv2
import deeplabcut
import imagehash
import pandas as pd
from PIL import Image
from ruamel.yaml import YAML

from .xma_data_processor import XMADataProcessor
from .xrommtools import analyze_xromm_videos


class Analyzer:
    """Analyzes XROMM videos using trained network."""

    def __init__(self, config):
        self._trials_path = os.path.join(config["working_dir"], "trials")
        self._data_processor = XMADataProcessor(config)
        self._config = config
        self._dlc_config = config['path_config_file']

    def analyze_videos(self):
        '''Analyze videos with a pre-existing network'''
        trials = self._data_processor.list_trials()

        # Establish project vars
        yaml = YAML()
        with open(self._dlc_config) as dlc_config:
            dlc = yaml.load(dlc_config)
        iteration = dlc['iteration']

        mode = self._config['tracking_mode']
        if mode == '2D':
            analyze_xromm_videos(self._dlc_config, self._trials_path, iteration)
        elif mode == 'per_cam':
            try:
                analyze_xromm_videos(path_config_file=self._dlc_config,
                                    path_config_file_cam2=self._config['path_config_file_2'],
                                    path_data_to_analyze=self._trials_path,
                                    iteration=iteration,
                                    nnetworks=2)
            except KeyError as e:
                print("Path to second DLC config not found. Did you create the project as a per-cam project?")
                print("If not, re-run 'create_new_project' using mode='per_cam'")
                raise e

        else:
            for trial in trials:
                trial_path =  os.path.join(self._trials_path, trial)
                video_path = f'{trial_path}/{trial}_rgb.avi'
                if not os.path.exists(video_path):
                    self._data_processor.make_rgb_video(trial_path)
                destfolder = f'{trial_path}/it{iteration}/'
                deeplabcut.analyze_videos(self._dlc_config, video_path, destfolder=destfolder, save_as_csv=True)
                self._split_dlc_to_xma(trial)

    def analyze_video_similarity_project(self):
        '''Analyze all videos in a project and take their average similar. This is dangerous, as it will assume that all cam1/cam2 pairs match
        or don't match!'''
        similarity_score = {}
        yaml = YAML()
        trial_perms = combinations(self._data_processor.list_trials(), 2)
        for trial1, trial2 in trial_perms:
            self._config['trial_1_name'] = trial1
            self._config['trial_2_name'] = trial2
            with open(os.path.join(self._config['working_dir'], 'project_config.yaml'), 'w') as file:
                yaml.dump(self._config, file)
            # TODO: pass the trial names directly instead of writing into config
            similarity_score[(trial1, trial2)] = self.analyze_video_similarity_trial()
        return similarity_score

    def analyze_video_similarity_trial(self):
        '''Analyze the average similarity between trials using image hashing'''
        trials = [self._config['trial_1_name'], self._config['trial_2_name']]
        cameras = ['cam1', 'cam2']
        videos = {
            (trial, cam): cv2.VideoCapture(
                self._data_processor.find_cam_file(os.path.join(self._trials_path, trial), cam))
            for trial in trials for cam in cameras
        }

        # Compare hashes based on the camera views configuration
        if self._config['cam1s_are_the_same_view']:
            cam1_diff, noc1 = self._compare_two_videos(videos[(trials[0], 'cam1')], videos[(trials[1], 'cam1')])
            cam2_diff, noc2 = self._compare_two_videos(videos[(trials[0], 'cam2')], videos[(trials[1], 'cam2')])
        else:
            cam1_diff, noc1 = self._compare_two_videos(videos[(trials[0], 'cam1')], videos[(trials[1], 'cam2')])
            cam2_diff, noc2 = self._compare_two_videos(videos[(trials[0], 'cam2')], videos[(trials[1], 'cam1')])

        # Calculate the average difference for each camera view
        cam1_avg_diff = cam1_diff / noc1 if noc1 > 0 else 0
        cam2_avg_diff = cam2_diff / noc2 if noc2 > 0 else 0

        # Calculate the overall trial average difference
        trial_avg_diff = (cam1_avg_diff + cam2_avg_diff) / 2

        # Note: The number of comparisons (noc) grows with video size, which could potentially affect the similarity measure,
        # making larger videos appear more similar than they actually are. This aspect may need further consideration.
        return trial_avg_diff

    def analyze_marker_similarity_project(self):
        '''Analyze all videos in a project and get their average rhythmicity. This assumes that all cam1/2 pairs are either the same or different!'''
        marker_similarity = {}
        yaml = YAML()

        trial_perms = combinations(self._data_processor.list_trials(), 2)
        for trial1, trial2 in trial_perms:
            self._config['trial_1_name'] = trial1
            self._config['trial_2_name'] = trial2
            with open(os.path.join(self._config['working_dir'], 'project_config.yaml'), 'w') as file:
                yaml.dump(self._config, file)
            # TODO: pass the trial names directly instead of writing into config
            marker_similarity[(trial1, trial2)] = abs(self.analyze_marker_similarity_trial())
        return marker_similarity

    def analyze_marker_similarity_trial(self):
        '''Analyze marker similarity for a pair of trials using the distance formula.'''
        # Find CSVs for each trial
        trial1 = self._config['trial_1_name']
        trial2 = self._config['trial_2_name']
        trial1_path = os.path.join(self._trials_path, trial1)
        trial2_path = os.path.join(self._trials_path, trial2)

        # Get a list of markers that each trial have in common
        bodyparts1 = self._data_processor.get_bodyparts_from_xma(trial1_path, mode='rgb')
        bodyparts2 = self._data_processor.get_bodyparts_from_xma(trial2_path, mode='rgb')
        markers_in_common = [marker for marker in bodyparts1 if marker in bodyparts2]

        trial1_csv = pd.read_csv(os.path.join(trial1_path, f'{trial1}.csv'))
        trial2_csv = pd.read_csv(os.path.join(trial2_path, f'{trial2}.csv'))

        avg_distances = []
        for marker in markers_in_common:
            avg_x1, avg_y1 = trial1_csv[f'{marker}_X'].mean(), trial1_csv[f'{marker}_Y'].mean()
            avg_x2, avg_y2 = trial2_csv[f'{marker}_X'].mean(), trial2_csv[f'{marker}_Y'].mean()

            # Calculate the distance between the average positions for this marker in the two trials
            distance = math.sqrt((avg_x2 - avg_x1) ** 2 + (avg_y2 - avg_y1) ** 2)
            avg_distances.append(distance)

        # Calculate the mean of the distances to get an overall similarity measure
        marker_similarity = sum(avg_distances) / len(avg_distances) if avg_distances else 0

        return marker_similarity

    def get_max_dissimilarity_for_trial(self, trial_path, window):
        '''Calculate the dissimilarity within the trial given the frame sliding window.'''
        video1 = cv2.VideoCapture(self._data_processor.find_cam_file(trial_path, 'cam1'))
        video2 = cv2.VideoCapture(self._data_processor.find_cam_file(trial_path, 'cam2'))

        hashes1 = self._hash_trial_video(video1)
        hashes2 = self._hash_trial_video(video2)

        return self._find_dissimilar_regions(hashes1, hashes2, window)

    def _split_dlc_to_xma(self, trial, save_hdf=True):
        '''Takes the output from RGB deeplabcut and splits it into XMAlab-readable output'''
        bodyparts_xy = []
        yaml = YAML()
        with open(self._dlc_config) as dlc_config:
            dlc = yaml.load(dlc_config)
        iteration = dlc['iteration']
        trial_path = os.path.join(self._trials_path, trial)

        rgb_parts = self._data_processor.get_bodyparts_from_xma(trial_path, mode='rgb')
        for part in rgb_parts:
            bodyparts_xy.append(part+'_X')
            bodyparts_xy.append(part+'_Y')

        csv_path = [file for file in os.listdir(f'{trial_path}/it{iteration}') if '.csv' in file and '-2DPoints' not in file]
        if len(csv_path) > 1:
            raise FileExistsError('Found more than 1 data CSV for RGB trial. Please remove CSVs from older analyses from this folder before analyzing.')
        if len(csv_path) < 1:
            raise FileNotFoundError(f'Couldn\'t find data CSV for trial {trial}. Something wrong with DeepLabCut?')

        csv_path = csv_path[0]
        xma_csv_path = f'{trial_path}/it{iteration}/{trial}-Predicted2DPoints.csv'

        df = pd.read_csv(f'{trial_path}/it{iteration}/{csv_path}', skiprows=1)
        df.index = df['bodyparts']
        df = df.drop(columns=df.columns[df.loc['coords'] == 'likelihood'])
        df = df.drop(columns=[column for column in df.columns if column not in rgb_parts and column not in [f'{bodypart}.1' for bodypart in rgb_parts]])
        df.columns = bodyparts_xy
        df = df.drop(index='coords')
        df.to_csv(xma_csv_path, index=False)
        print("Successfully split DLC format to XMALab 2D points; saved "+str(xma_csv_path))
        if save_hdf:
            tracked_hdf = os.path.splitext(csv_path)[0]+'.h5'
            df.to_hdf(tracked_hdf, 'df_with_missing', format='table', mode='w', nan_rep='NaN')

    def _compare_two_videos(self, video1, video2):
        '''Compare two videos using image hashing'''
        hashes1 = self._hash_trial_video(video1)
        hashes2 = self._hash_trial_video(video2)

        video1_frames = len(hashes1)
        video2_frames = len(hashes2)
        noc = math.perm(video1_frames + video2_frames, 2)  # Might need revision based on actual comparison logic

        print(f'Video 1 frames: {video1_frames}')
        print(f'Video 2 frames: {video2_frames}')

        print('Comparing hashes between videos')
        hash_dif = sum(hash1 - hash2 for hash1 in hashes1 for hash2 in hashes2)

        return hash_dif, noc

    def _hash_trial_video(self, video):
        '''Generate image hashes for a single video'''
        video_frames = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
        print(f'Video frames: {video_frames}')

        hashes = []
        print('Creating hashes for video')
        for i in range(video_frames):
            print(f'Current frame: {i}')
            ret, frame = video.read()
            if not ret:
                print('Error reading video frame')
                cv2.destroyAllWindows()
                break
            hashes.append(imagehash.phash(Image.fromarray(frame)))

        return hashes

    def _find_dissimilar_regions(self, hashes1, hashes2, window):
        '''Find the region of maximum dissimilarity given 2 lists of hashes and a sliding window (how many frames)'''
        start_frame_vid1 = 0
        start_frame_vid2 = 0
        max_hash_dif_vid1 = 0
        max_hash_dif_vid2 = 0
        hash_dif_vid1 = 0
        hash_dif_vid2 = 0

        for slider in range(0, len(hashes1) // window):
            print(f'Current start frame {slider * window}')
            hash_dif_vid1, hash_dif_vid2 = self._compare_hash_sets(hashes1[slider * window:(slider + 1) * window], hashes2[slider * window:(slider + 1) * window])

            print(f'Current hash diff (vid 1): {hash_dif_vid1}')
            print(f'Current hash diff (vid 2): {hash_dif_vid2}')
            if hash_dif_vid1 > max_hash_dif_vid1:
                max_hash_dif_vid1 = hash_dif_vid1
                start_frame_vid1 = slider * window

            if hash_dif_vid2 > max_hash_dif_vid2:
                max_hash_dif_vid2 = hash_dif_vid2
                start_frame_vid2 = slider * window

            print(f'Max hash diff (vid 1): {max_hash_dif_vid1}')
            print(f'Max hash diff (vid 2): {max_hash_dif_vid2}')

            print(f'Start frame (vid 1): {start_frame_vid1}')
            print(f'Start frame (vid 2): {start_frame_vid2}')

        return start_frame_vid1, start_frame_vid2

    def _compare_hash_sets(self, hashes1, hashes2):
        '''Compares two sets of image hashes to find dissimilarities'''
        hash1_dif = 0
        hash2_dif = 0

        print(f'Hash set 1 {hashes1[0]}')
        print(f'Hash set 2 {hashes2[0]}')
        # Compares all possible combinations of images
        for combination in combinations(hashes1, 2):
            hash1_dif = hash1_dif + (combination[0] - combination[1])

        for combination in combinations(hashes2, 2):
            hash2_dif = hash2_dif + (combination[0] - combination[1])

        return hash1_dif, hash2_dif



"""
XROMM Tools for DeepLabCut
Developed by J.D. Laurence-Chasen

Functions:

xma_to_dlc: create DeepLabCut training dataset from data tracked in XMALab
analyze_xromm_videos: Predict 2D points for novel trials
dlc_to_xma: convert output of DeepLabCut to XMALab format 2D points file
add_frames: Add new frames corrected/tracked in XMALab to an existing training dataset

"""


import os
import pandas as pd
import numpy as np
import cv2
import random
from deeplabcut.pose_estimation_tensorflow.predict_videos import analyze_videos

def xma_to_dlc(path_config_file,data_path,dataset_name,scorer,nframes,nnetworks = 1, path_config_file_cam2 = []):
    config = path_config_file[:-12]
    cameras = [1,2]
    picked_frames = []
    dfs = []
    idx = []
    pnames = []
    subs =[["c01","c1","C01","C1","Cam1","cam1","Cam01","cam01","Camera1","camera1"],["c02","c2","C02","C2","Cam2","cam2","Cam02","cam02","Camera2","camera2"]]
    trialnames = [folder for folder in os.listdir(data_path) if os.path.isdir(os.path.join(data_path, folder)) and not folder.startswith('.')]


    ### PART 1: Pick frames for dataset

    for trial in trialnames:

        # Read 2D points file
        contents = os.listdir(data_path+"/"+trial)
        filename = [x for x in contents if ".csv" in x] # csv filename
        df1=pd.read_csv(data_path+"/"+trial+"/"+filename[0], sep=',',header=None)

        # read pointnames from header row
        pointnames = df1.loc[0,::4].astype(str).str[:-7].tolist()
        pnames.append(pointnames)

        df1 = df1.loc[1:,].reset_index(drop=True) # remove header row

        # temp_idx = rows where fewer than half of columns are NaN
        ncol = df1.shape[1]
        temp_idx = list(df1.index.values[(~pd.isnull(df1)).sum(axis = 1) >= ncol/2])


        # randomize frames w/i each trial and append to index list
        random.shuffle(temp_idx)
        idx.append(temp_idx)
        dfs.append(df1)


    # a couple errors
    if sum(len(x) for x in idx) < nframes:
        raise ValueError('nframes is bigger than number of detected frames')

    #if pointnames aren't the same across trials
    if any(pnames[0] != x for x in pnames):
        raise ValueError('Make sure point names are consistent across trials')

    # pick frames to extract (NOTE this is random currently)
    # current code iteratively picks one frame at a time from each shuffled trial until # of picked_frames hits nframes
    # There is a much neater way to do this
    count = 0
    while sum(len(x) for x in picked_frames) < nframes:
        for trialnum in range(len(idx)):
            if sum(len(x) for x in picked_frames) < nframes:
                if count == 0:
                    picked_frames.insert(trialnum,[idx[trialnum][count]])
                elif count < len(idx[trialnum]):
                    picked_frames[trialnum] = picked_frames[trialnum]+[idx[trialnum][count]]
        count += 1

    ### Part 2: Extract images and 2D point data

    if nnetworks == 2:
        configs = [path_config_file[:-12], path_config_file_cam2[:-12]]

        for camera in cameras:
            print("Extracting camera %d trial images and 2D points..."%camera)
            relnames = []
            data = pd.DataFrame()
            # new training dataset folder
            newpath = configs[camera-1]+"/labeled-data/"+dataset_name+"_cam"+str(camera)
            h5_save_path = newpath+"/CollectedData_"+scorer+".h5"
            csv_save_path = newpath+"/CollectedData_"+scorer+".csv"

            if os.path.exists(newpath):
                contents = os.listdir(newpath)
                if contents:
                    raise ValueError('There are already data in the camera %d training dataset folder' %camera)
            else:
                os.makedirs(newpath) # make new folder

            for trialnum,trial in enumerate(trialnames):

                # get video file
                file = []
                contents = os.listdir(data_path+"/"+trial)
                for name in contents:
                    if any(x in name for x in subs[camera-1]):
                        file = name
                if not file:
                    raise ValueError('Cannot locate %s video file or image folder' %trial)

                # if video file is actually folder of frames
                if os.path.isdir(data_path+"/"+trial+"/"+file):
                    imgpath = data_path+"/"+trial+"/"+file
                    imgs = os.listdir(imgpath)
                    relpath = "labeled-data/"+dataset_name+"_cam"+str(camera)+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()

                    for count,img in enumerate(imgs):
                        if count in frames:
                            image = cv2.imread(imgpath+"/"+img)
                            relname = relpath + trial + "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(newpath + "/" + trial + "_%s.png" % str(count+1).zfill(4), image) # save frame


                else:
                # file is actually a file
                # extract frames from video and convert to png
                    video = data_path+"/"+trial+"/"+file
                    relpath = "labeled-data/"+dataset_name+"_cam"+str(camera)+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()
                    cap = cv2.VideoCapture(video)
                    success,image = cap.read()
                    count = 0
                    while success:
                        if count in frames:
                            relname = relpath + trial + "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(newpath + "/" + trial + "_%s.png" % str(count+1).zfill(4), image) # save frame
                        success,image = cap.read()
                        count += 1
                    cap.release()

                # extract 2D points data
                df1= dfs[trialnum]
                xpos = df1.iloc[frames,0+(camera-1)*2::4]
                ypos = df1.iloc[frames,1+(camera-1)*2::4]
                temp_data = pd.concat([xpos,ypos],axis=1).sort_index(axis=1)
                data = pd.concat([data,temp_data])

        ### Part 3: Complete final structure of datafiles
            dataFrame = pd.DataFrame()
            temp = np.empty((data.shape[0],2,))
            temp[:] = np.nan
            for i,bodypart in enumerate(pointnames):
                index = pd.MultiIndex.from_product([[scorer], [bodypart], ['x', 'y']],names=['scorer', 'bodyparts', 'coords'])
                frame = pd.DataFrame(temp, columns = index, index = relnames)
                frame.iloc[:,0:2] = data.iloc[:, 2*i:2*i+2].values.astype(float)
                dataFrame = pd.concat([dataFrame, frame],axis=1)
            dataFrame.replace('', np.nan, inplace=True)
            dataFrame.replace(' NaN', np.nan, inplace=True)
            dataFrame.replace(' NaN ', np.nan, inplace=True)
            dataFrame.replace('NaN ', np.nan, inplace=True)
            dataFrame.apply(pd.to_numeric)
            dataFrame.to_hdf(h5_save_path, key="df_with_missing", mode="w")
            dataFrame.to_csv(csv_save_path,na_rep='NaN')
            print("...done.")

    else:

        relnames = []
        data = pd.DataFrame()
        # new training dataset folder
        newpath = config+"/labeled-data/"+dataset_name
        h5_save_path = newpath+"/CollectedData_"+scorer+".h5"
        csv_save_path = newpath+"/CollectedData_"+scorer+".csv"

        if os.path.exists(newpath):
            contents = os.listdir(newpath)
            if contents:
                raise ValueError('There are already data in the camera %d training dataset folder' %camera)
        else:
            os.makedirs(newpath) # make new folder

        for camera in cameras:
            print("Extracting camera %d trial images and 2D points..."%camera)

            for trialnum,trial in enumerate(trialnames):

                # get video file
                file = []
                contents = os.listdir(data_path+"/"+trial)
                for name in contents:
                    if any(x in name for x in subs[camera-1]):
                        file = name
                if not file:
                    raise ValueError('Cannot locate %s video file or image folder' %trial)

                # if video file is actually folder of frames
                if os.path.isdir(data_path+"/"+trial+"/"+file):
                    imgpath = data_path+"/"+trial+"/"+file
                    imgs = os.listdir(imgpath)
                    relpath = "labeled-data/"+dataset_name+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()

                    for count,img in enumerate(imgs):
                        if count in frames:
                            image = cv2.imread(imgpath+"/"+img)
                            relname = relpath + trial + "_cam"+str(camera)+ "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(newpath + "/" + trial + "_cam"+str(camera)+ "_%s.png" % str(count+1).zfill(4), image) # save frame


                else:
                # file is actually a file
                # extract frames from video and convert to png
                    video = data_path+"/"+trial+"/"+file
                    relpath = "labeled-data/"+dataset_name+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()
                    cap = cv2.VideoCapture(video)
                    success,image = cap.read()
                    count = 0
                    while success:
                        if count in frames:
                            relname = relpath + trial + "_cam"+str(camera)+ "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(newpath + "/" + trial + "_cam"+str(camera)+ "_%s.png" % str(count+1).zfill(4), image) # save frame
                        success,image = cap.read()
                        count += 1
                    cap.release()

                # extract 2D points data
                df1= dfs[trialnum]
                xpos = df1.iloc[frames,0+(camera-1)*2::4]
                ypos = df1.iloc[frames,1+(camera-1)*2::4]
                temp_data = pd.concat([xpos,ypos],axis=1).sort_index(axis=1)
                temp_data.columns = range(temp_data.shape[1])
                data = pd.concat([data,temp_data])

    ### Part 3: Complete final structure of datafiles
        dataFrame = pd.DataFrame()
        temp = np.empty((data.shape[0],2,))
        temp[:] = np.nan
        for i,bodypart in enumerate(pointnames):
            index = pd.MultiIndex.from_product([[scorer], [bodypart], ['x', 'y']],names=['scorer', 'bodyparts', 'coords'])
            frame = pd.DataFrame(temp, columns = index, index = relnames)
            frame.iloc[:,0:2] = data.iloc[:, 2*i:2*i+2].values.astype(float)
            dataFrame = pd.concat([dataFrame, frame],axis=1)
        dataFrame.replace('', np.nan, inplace=True)
        dataFrame.replace(' NaN', np.nan, inplace=True)
        dataFrame.replace(' NaN ', np.nan, inplace=True)
        dataFrame.replace('NaN ', np.nan, inplace=True)
        dataFrame.apply(pd.to_numeric)
        dataFrame.to_hdf(h5_save_path, key="df_with_missing", mode="w")
        dataFrame.to_csv(csv_save_path,na_rep='NaN')
        print("...done.")

    print("Training data extracted to projectpath/labeled-data. Now use deeplabcut.create_training_dataset")

def dlc_to_xma(cam1data,cam2data,trialname,savepath):

    h5_save_path = savepath+"/"+trialname+"-Predicted2DPoints.h5"
    csv_save_path = savepath+"/"+trialname+"-Predicted2DPoints.csv"

    if isinstance(cam1data, str): #is string
        if ".csv" in cam1data:

            cam1data=pd.read_csv(cam1data, sep=',',header=None)
            cam2data=pd.read_csv(cam2data, sep=',',header=None)
            pointnames = list(cam1data.loc[1,1:].unique())

            # reformat CSV / get rid of headers
            cam1data = cam1data.loc[3:,1:]
            cam1data.columns = range(cam1data.shape[1])
            cam1data.index = range(cam1data.shape[0])
            cam2data = cam2data.loc[3:,1:]
            cam2data.columns = range(cam2data.shape[1])
            cam2data.index = range(cam2data.shape[0])

        elif ".h5" in cam1data:# is .h5 file
            cam1data = pd.read_hdf(cam1data)
            cam2data = pd.read_hdf(cam2data)
            pointnames = list(cam1data.columns.get_level_values('bodyparts').unique())

        else:
            raise ValueError('2D point input is not in correct format')
    else:

        pointnames = list(cam1data.columns.get_level_values('bodyparts').unique())

    # make new column names
    nvar = len(pointnames)
    pointnames = [item for item in pointnames for repetitions in range(4)]
    post = ["_cam1_X", "_cam1_Y", "_cam2_X", "_cam2_Y"]*nvar
    cols = [m+str(n) for m,n in zip(pointnames,post)]


    # remove likelihood columns
    cam1data = cam1data.drop(cam1data.columns[2::3],axis=1)
    cam2data = cam2data.drop(cam2data.columns[2::3],axis=1)

    # replace col names with new indices
    c1cols = list(range(0,cam1data.shape[1]*2,4)) + list(range(1,cam1data.shape[1]*2,4))
    c2cols = list(range(2,cam1data.shape[1]*2,4)) + list(range(3,cam1data.shape[1]*2,4))
    c1cols.sort()
    c2cols.sort()
    cam1data.columns = c1cols
    cam2data.columns = c2cols

    df = pd.concat([cam1data,cam2data],axis=1).sort_index(axis=1)
    df.columns = cols
    df.to_hdf(h5_save_path, key="df_with_missing", mode="w")
    df.to_csv(csv_save_path,na_rep='NaN',index=False)

def analyze_xromm_videos(path_config_file,path_data_to_analyze,iteration,nnetworks = 1, path_config_file_cam2 = []):
# assumes you have cam1 and cam2 videos as .avi in their own seperate trial folders
# assumes all folders w/i new_data_path are trial folders
# convert jpg stacks?

# analyze videos
    cameras = [1,2]
    config = path_config_file
    configs = [path_config_file, path_config_file_cam2]
    subs =[["c01","c1","C01","C1","Cam1","cam1","Cam01","cam01","Camera1","camera1"],["c02","c2","C02","C2","Cam2","cam2","Cam02","cam02","Camera2","camera2"]]
    trialnames = [folder for folder in os.listdir(path_data_to_analyze) if os.path.isdir(os.path.join(path_data_to_analyze, folder)) and not folder.startswith('.')]

    for trialnum,trial in enumerate(trialnames):
        trialpath = os.path.join(path_data_to_analyze,trial)
        contents = os.listdir(trialpath)
        savepath = os.path.join(trialpath, f"it{iteration}")
        if os.path.exists(savepath):
            temp = os.listdir(savepath)
            if temp:
                raise ValueError('There are already predicted points in iteration %d subfolders' %iteration)
        else:
            os.makedirs(savepath) # make new folder
        # get video file
        for camera in cameras:
            for name in contents:
                if any(x in name for x in subs[camera-1]):
                    filename = name
            if not filename:
                raise ValueError('Cannot locate %s video file or image folder' %trial)

            video = os.path.join(trialpath, filename)
            print(video)
            #analyze video
            if nnetworks == 1:
                analyze_videos(config,[video],destfolder = savepath,save_as_csv=True)
            else:
                analyze_videos(configs[camera-1],[video],destfolder = savepath,save_as_csv=True)

        # get filenames and read analyzed data
        contents = os.listdir(savepath)
        datafiles = [s for s in contents if '.h5' in s]
        if not datafiles:
            raise ValueError('Cannot find predicted points. Some wrong with DeepLabCut?')
        cam1data = pd.read_hdf(savepath+"/"+datafiles[0])
        cam2data = pd.read_hdf(savepath+"/"+datafiles[1])
        dlc_to_xma(cam1data,cam2data,trial,savepath)

def add_frames(path_config_file, data_path, iteration, frames, nnetworks = 1, path_config_file_cam2 = "enterpathofcam2config"):

    #input: config file paths, path of data to add to trainingdataset, frames-csv file where first col is trialnames and following cols are frame numbers
    # will look for 2D points file based on name (if there are multiple csv files)

    configs = [path_config_file[:-12], path_config_file_cam2[:-12]]
    cameras = [1,2]
    subs =[["c01","c1","C01","C1","Cam1","cam1","Cam01","cam01","Camera1","camera1"],["c02","c2","C02","C2","Cam2","cam2","Cam02","cam02","Camera2","camera2"]]
    pts = ["2Dpts","2dpts","2DPts","2dPts","pts2D","Pts2D","pts2d","points2D","Points2d","points2d","2Dpoints","2dpoints","2DPoints"]
    corr = ["correct","Correct","corrected","Corrected"]
    # read frames from csv
    if '.csv' in frames:
        f = pd.read_csv(frames,header=None)
        trialnames = list(f.iloc[:,0]) # first row of frames file must be trialnames
        picked_frames = []
        # this is disgusting code
        for row in range(f.shape[0]):
            picked_frames.append(list(f.loc[row,1:]))
        for count,row in enumerate(picked_frames):
            picked_frames[count] = [x for x in row if str(x) != 'nan'] # remove nans
        for count,row in enumerate(picked_frames):
            picked_frames[count] = [int(x) for x in row] # convert to int
    else:
        raise ValueError('frames must be a .csv file with trialnames and frame numbers')

    if nnetworks == 2:

        for camera in cameras:
            contents = os.listdir(configs[camera-1]+'/'+'labeled-data')
            if len(contents) == 1:
                dataset_name = contents[0]
                labeleddata_path = configs[camera-1] + '/' + 'labeled-data/' + dataset_name
            else:
                raise ValueError('There must be only one data set in the labeled-data folder')

            contents = os.listdir(labeleddata_path)
            h5file = [x for x in contents if '.h5' in x]
            csvfile = [x for x in contents if '.csv' in x]
            data = pd.read_hdf(labeleddata_path+'/'+h5file[0]) # read old point labels

            ## Extract selected frames from videos

            for trialnum,trial in enumerate(trialnames):
            # get video file
                file = []
                relnames = []
                contents = os.listdir(data_path+"/"+trial)
                for name in contents:
                    if any(x in name for x in subs[camera-1]):
                        file = name
                if not file:
                    raise ValueError('Cannot locate %s video file or image folder' %trial)

                # if video file is actually folder of frames
                if os.path.isdir(data_path+"/"+trial+"/"+file):
                    imgpath = data_path+"/"+trial+"/"+file
                    imgs = os.listdir(imgpath)
                    relpath = "labeled-data/"+dataset_name+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()

                    for count,img in enumerate(imgs):
                        if count+1 in frames: # ASSUMES FRAMES PROVIDED ARE 1 index
                            image = cv2.imread(imgpath+"/"+img)
                            relname = relpath + trial + "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(labeleddata_path + "/" + trial + "_%s.png" % str(count+1).zfill(4), image) # save frame
                else:
                # file is actually a file
                # extract frames from video and convert to png
                    video = data_path+"/"+trial+"/"+file
                    relpath = "labeled-data/"+dataset_name+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()
                    cap = cv2.VideoCapture(video)
                    success,image = cap.read()
                    count = 0
                    while success:
                        if count+1 in frames:
                            relname = relpath + trial + "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(labeleddata_path + "/" + trial + "_%s.png" % str(count+1).zfill(4), image) # save frame
                        success,image = cap.read()
                        count += 1
                    cap.release()

                # get 2D points file / data
                # extract 2D points data
                contents = os.listdir(data_path+"/"+trial+"/"+"it"+str(iteration))
                pointsfile = [x for x in contents if '.csv' in x]

                if not pointsfile:
                    raise ValueError('Cannot locate %s 2D points file' %trial)

                # if multiple csv files, look for "2Dpoints" in the name
                if len(pointsfile) > 1:
                    t= []
                    for q in pointsfile:
                        if any(x in q for x in pts):
                            t = t + [q]
                    # if there are multiple 2D points files, look for "corrected" in the name
                    if len(t) > 1:
                        for r in pointsfile:
                            if any(x in r for x in corr):
                                file = r
                    else:
                        file = t[0]
                else:
                    file = pointsfile[0]
                print("Reading and adding the following frames from " + data_path+'/'+trial+"/"+"it"+str(iteration)+'/'+file)
                df = pd.read_csv(data_path+'/'+trial+"/"+"it"+str(iteration)+'/'+file,sep=',',header=None)
                df = df.loc[1:,].reset_index(drop=True)
                print(frames)
                frames = [x - 1 for x in frames] # account for zero index in python
                xpos = df.iloc[frames,0+(camera-1)*2::4]
                ypos = df.iloc[frames,1+(camera-1)*2::4]
                temp_data = pd.concat([xpos,ypos],axis=1).sort_index(axis=1)
                if temp_data.shape[1] > data.shape[1]:
                    raise ValueError('There are %d extra points in the corrected points file'%((temp_data.shape[1] - data.shape[1])/2))
                if temp_data.shape[1] < data.shape[1]:
                    raise ValueError('There are %d missing points in the corrected points file'%((data.shape[1] - temp_data.shape[1])/2))
                temp_data.index = relnames
                temp_data.columns = data.columns
                data = pd.concat([data,temp_data])
            data.replace(' NaN', np.nan, inplace=True)
            data.replace(' NaN ', np.nan, inplace=True)
            data.replace('NaN ', np.nan, inplace=True)
            data = data.astype('float')
            data = data.round(2)
            data = data.apply(pd.to_numeric)
            data.to_hdf(labeleddata_path+'/'+h5file[0], key="df_with_missing", mode="w")
            data.to_csv(labeleddata_path+'/'+csvfile[0],na_rep='NaN')

    else: # default, one network for both videos
        config = path_config_file[:-12]
        contents = os.listdir(config+'/'+'labeled-data')
        if len(contents) == 1:
            dataset_name = contents[0]
            labeleddata_path = config + '/' + 'labeled-data/' + dataset_name
        else:
            raise ValueError('There must be only one data set in the labeled-data folder')

        contents = os.listdir(labeleddata_path)
        h5file = [x for x in contents if '.h5' in x]
        csvfile = [x for x in contents if '.csv' in x]
        data = pd.read_hdf(labeleddata_path+'/'+h5file[0]) # read old point labels

        for camera in cameras:

        ## Extract selected frames from videos

            for trialnum,trial in enumerate(trialnames):
            # get video file
                relnames = []
                file = []

                contents = os.listdir(data_path+"/"+trial)
                for name in contents:
                    if any(x in name for x in subs[camera-1]):
                        file = name
                if not file:
                    raise ValueError('Cannot locate %s video file or image folder' %trial)

                # if video file is actually folder of frames
                if os.path.isdir(data_path+"/"+trial+"/"+file):
                    imgpath = data_path+"/"+trial+"/"+file
                    imgs = os.listdir(imgpath)
                    relpath = "labeled-data/"+dataset_name+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()

                    for count,img in enumerate(imgs):
                        if count+1 in frames: # ASSUMES FRAMES PROVIDED ARE 1 index
                            image = cv2.imread(imgpath+"/"+img)
                            relname = relpath + trial + "_cam"+str(camera) +"_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(labeleddata_path + "/" + trial + "_cam"+str(camera) + "_%s.png" % str(count+1).zfill(4), image) # save frame
                else:
                # file is actually a file
                # extract frames from video and convert to png
                    video = data_path+"/"+trial+"/"+file
                    relpath = "labeled-data/"+dataset_name+"/"
                    frames = picked_frames[trialnum]
                    frames.sort()
                    cap = cv2.VideoCapture(video)
                    success,image = cap.read()
                    count = 0
                    while success:
                        if count+1 in frames:
                            relname = relpath + trial + "_cam"+str(camera) + "_%s.png" % str(count+1).zfill(4)
                            relnames = relnames + [relname]
                            cv2.imwrite(labeleddata_path + "/" + trial + "_cam"+str(camera)+ "_%s.png" % str(count+1).zfill(4), image) # save frame
                        success,image = cap.read()
                        count += 1
                    cap.release()

                # get 2D points file / data
                # extract 2D points data
                contents = os.listdir(data_path+"/"+trial+"/"+"it"+str(iteration))
                pointsfile = [x for x in contents if '.csv' in x]

                if not pointsfile:
                    raise ValueError('Cannot locate %s 2D points file' %trial)

                # if multiple csv files, look for "2Dpoints" in the name
                if len(pointsfile) > 1:
                    t= []
                    for q in pointsfile:
                        if any(x in q for x in pts):
                            t = t + [q]
                    # if there are multiple 2D points files, look for "corrected" in the name
                    if len(t) > 1:
                        for r in pointsfile:
                            if any(x in r for x in corr):
                                pointsfile = r
                else:
                    pointsfile = pointsfile[0]
                if isinstance(pointsfile,str) != True:
                    raise ValueError('Please check the points files in trial '+trial+' iteration '+str(iteration)+' folder')
                df = pd.read_csv(data_path+'/'+trial+"/"+"it"+str(iteration)+'/'+pointsfile,sep=',',header=None)
                df = df.loc[1:,].reset_index(drop=True)
                xpos = df.iloc[frames,0+(camera-1)*2::4]
                ypos = df.iloc[frames,1+(camera-1)*2::4]
                temp_data = pd.concat([xpos,ypos],axis=1).sort_index(axis=1)
                temp_data.index = relnames
                if temp_data.shape[1] > data.shape[1]:
                    raise ValueError('There are %d extra points in the corrected points file'%((temp_data.shape[1] - data.shape[1])/2))
                if temp_data.shape[1] < data.shape[1]:
                    raise ValueError('There are %d missing points in the corrected points file'%((data.shape[1] - temp_data.shape[1])/2))
                temp_data.columns = data.columns
                data = pd.concat([data,temp_data])
        data.replace(' NaN', np.nan, inplace=True)
        data.replace(' NaN ', np.nan, inplace=True)
        data.replace('NaN ', np.nan, inplace=True)
        data = data.astype('float')
        data = data.round(2)
        data = data.apply(pd.to_numeric)
        data.to_hdf(labeleddata_path+'/'+h5file[0], key="df_with_missing", mode="w")
        data.to_csv(labeleddata_path+'/'+csvfile[0],na_rep='NaN')

    print('Frames from %d trials successfully added to training dataset'%len(trialnames))



"""Performs XMAlab-style autocorrection on the trials and videos"""

import math
import os

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from ruamel.yaml import YAML

from .xma_data_processor import XMADataProcessor


class Autocorrector:
    """Performs XMAlab-style autocorrection on the trials and videos"""

    def __init__(self, config):
        self._trials_path = os.path.join(config["working_dir"], "trials")
        self._data_processor = XMADataProcessor(config)
        self._config = config
        self._dlc_config = config['path_config_file']

    def autocorrect_trial(self):
        '''Do XMAlab-style autocorrect on the tracked beads'''
        trials = self._data_processor.list_trials()

        # Establish project vars
        yaml = YAML()
        with open(self._dlc_config) as dlc_config:
            dlc = yaml.load(dlc_config)

        iteration = dlc['iteration']

        for trial in trials:
            # Find the appropriate pointsfile
            trial_path = os.path.join(self._trials_path, trial)
            iteration_path = os.path.join(trial_path, f'it{iteration}')

            try:
                csv = pd.read_csv(os.path.join(iteration_path, f'{trial}-Predicted2DPoints.csv'))
            except FileNotFoundError as e:
                print(f'Could not find predicted 2D points file. Please check the it{iteration} folder for trial {trial}')
                raise e
            out_name = os.path.join(iteration_path, f'{trial}-AutoCorrected2DPoints.csv')

            if self._config['test_autocorrect']:
                cams = [self._config['cam']]
            else:
                cams = ['cam1', 'cam2']

            # For each camera
            for cam in cams:
                csv = self._autocorrect_video(cam, trial_path, csv)

            # Print when autocorrect finishes
            if not self._config['test_autocorrect']:
                print(f'Done! Saving to {out_name}')
                csv.to_csv(out_name, index=False)

    def _autocorrect_video(self, cam, trial_path, csv):
        '''Run the autocorrect function on a single video within a single trial'''
        # Find the raw video
        video = cv2.VideoCapture(self._data_processor.find_cam_file(trial_path, cam))
        if not video.isOpened():
            raise FileNotFoundError(f'Couldn\'t find a video at file path: {trial_path}') from None

        if self._config['test_autocorrect']:
            video.set(1, self._config['frame_num'] - 1)
            ret, frame = video.read()
            if ret is False:
                raise IOError('Error reading video frame')
            self._autocorrect_frame(trial_path, frame, cam, self._config['frame_num'], csv)
            return csv

        # For each frame of video
        print(f'Total frames in video: {video.get(cv2.CAP_PROP_FRAME_COUNT)}')

        for frame_index in range(int(video.get(cv2.CAP_PROP_FRAME_COUNT))):
            # Load frame
            print(f'Current Frame: {frame_index + 1}')
            ret, frame = video.read()
            if ret is False:
                raise IOError('Error reading video frame')
            csv = self._autocorrect_frame(trial_path, frame, cam, frame_index, csv)
        return csv

    def _autocorrect_frame(self, trial_path, frame, cam, frame_index, csv):
        '''Run the autocorrect function for a single frame (no output)'''
        # For each marker in the frame
        if self._config['test_autocorrect']:
            parts_unique = [self._config['marker']]
        else:
            parts_unique = self._data_processor.get_bodyparts_from_xma(trial_path, mode='2D')
        for part in parts_unique:
            # Find point and offsets
            x_float = csv.loc[frame_index, part + '_' + cam + '_X']
            y_float = csv.loc[frame_index, part + '_' + cam + '_Y']
            search_area_with_offset = self._config['search_area'] + 0.5
            x_start = int(x_float - search_area_with_offset)
            y_start = int(y_float - search_area_with_offset)
            x_end = int(x_float + search_area_with_offset)
            y_end = int(y_float + search_area_with_offset)

            subimage = frame[y_start:y_end, x_start:x_end]

            subimage_filtered = self._filter_image(
                subimage, self._config['krad'], self._config['gsigma'], self._config['img_wt'],
                self._config['blur_wt'], self._config['gamma'])

            subimage_float = subimage_filtered.astype(np.float32)
            radius = int(1.5 * 5 + 0.5) #5 might be too high
            sigma = radius * math.sqrt(2 * math.log(255)) - 1
            subimage_blurred = cv2.GaussianBlur(subimage_float, (2 * radius + 1, 2 * radius + 1), sigma)

            subimage_diff = subimage_float-subimage_blurred
            subimage_diff = cv2.normalize(subimage_diff, None, 0,255,cv2.NORM_MINMAX).astype(np.uint8)

            # Median
            subimage_median = cv2.medianBlur(subimage_diff, 3)

            # LUT
            subimage_median = self._filter_image(subimage_median, krad=3)

            # Thresholding
            subimage_median = cv2.cvtColor(subimage_median, cv2.COLOR_BGR2GRAY)
            min_val, _, _, _ = cv2.minMaxLoc(subimage_median)
            thres = 0.5 * min_val + 0.5 * np.mean(subimage_median) + self._config['threshold'] * 0.01 * 255
            _, subimage_threshold =  cv2.threshold(subimage_median, thres, 255, cv2.THRESH_BINARY_INV)

            # Gaussian blur
            subimage_gaussthresh = cv2.GaussianBlur(subimage_threshold, (3,3), 1.3)

            # Find contours
            contours, _ = cv2.findContours(subimage_gaussthresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE, offset=(x_start,y_start))
            contours_im = [contour-[x_start, y_start] for contour in contours]

            # Find closest contour
            dist = 1000
            best_index = -1
            detected_centers = {}
            for i, cnt in enumerate(contours):
                detected_center, _ = cv2.minEnclosingCircle(cnt)
                dist_tmp = math.sqrt((x_float - detected_center[0])**2 + (y_float - detected_center[1])**2)
                detected_centers[round(dist_tmp, 4)] = detected_center
                if dist_tmp < dist:
                    best_index = i
                    dist = dist_tmp

            if self._config['test_autocorrect']:
                print('Raw')
                self._show_crop(subimage, 15)

                print('Filtered')
                self._show_crop(subimage_filtered, 15)

                print(f'Blurred: {sigma}')
                self._show_crop(subimage_blurred, 15)

                print('Diff (Float - blurred)')
                self._show_crop(subimage_diff, 15)

                print('Median')
                self._show_crop(subimage_median, 15)

                print('Median filtered')
                self._show_crop(subimage_median, 15)

                print('Threshold')
                self._show_crop(subimage_threshold, 15)

                print('Gaussian')
                self._show_crop(subimage_threshold, 15)

                print('Best Contour')
                detected_center_im, _ = cv2.minEnclosingCircle(contours_im[best_index])
                self._show_crop(subimage, 15, contours = [contours_im[best_index]], detected_marker = detected_center_im)

            # Save center of closest contour to CSV
            if best_index >= 0:
                detected_center, _ = cv2.minEnclosingCircle(contours[best_index])
                csv.loc[frame_index, part + '_' + cam + '_X']  = detected_center[0]
                csv.loc[frame_index, part + '_' + cam + '_Y']  = detected_center[1]
        return csv

    def _filter_image(self, image, krad=17, gsigma=10, img_wt=3.6, blur_wt=-2.9, gamma=0.10):
        '''Filter the image to make it easier to see the bead'''
        krad = krad*2+1
        # Gaussian blur
        image_blur = cv2.GaussianBlur(image, (krad, krad), gsigma)
        # Add to original
        image_blend = cv2.addWeighted(image, img_wt, image_blur, blur_wt, 0)
        lut = np.array([((i/255.0)**gamma)*255.0 for i in range(256)])
        image_gamma = image_blend.copy()
        im_type = len(image_gamma.shape)
        if im_type == 2:
            image_gamma = lut[image_gamma]
        elif im_type == 3:
            image_gamma[:,:,0] = lut[image_gamma[:,:,0]]
            image_gamma[:,:,1] = lut[image_gamma[:,:,1]]
            image_gamma[:,:,2] = lut[image_gamma[:,:,2]]
        return image_gamma

    def _show_crop(self, src, center, scale=5, contours=None, detected_marker=None):
        '''Display a visual of the marker and Python's projected center'''
        if len(src.shape) < 3:
            src = cv2.cvtColor(src, cv2.COLOR_GRAY2BGR)
        image = src.copy().astype(np.uint8)
        image = cv2.resize(image, None, fx=scale, fy=scale, interpolation = cv2.INTER_CUBIC)
        if contours:
            overlay = image.copy()
            scaled_contours = [contour*scale for contour in contours]
            cv2.drawContours(overlay, scaled_contours, -1, (255,0,0),2)
            image = cv2.addWeighted(overlay, 0.25, image, 0.75, 0)
        cv2.drawMarker(image, (center*scale, center*scale), color = (0,255,255), markerType = cv2.MARKER_CROSS, markerSize = 10, thickness = 1)
        if detected_marker:
            cv2.drawMarker(image,
            (int(detected_marker[0]*scale),
            int(detected_marker[1]*scale)),
            color = (255,0,0),
            markerType = cv2.MARKER_CROSS,
            markerSize = 10,
            thickness = 1)
        plt.imshow(image)
        plt.show()



"""Primary interface for training the XROMM network using DLC"""

import os

import deeplabcut

from .xrommtools import xma_to_dlc
from .xma_data_processor import XMADataProcessor


class Network:
    """Trains an XROMM labeling network using DLC."""

    def __init__(self, config):
        self._data_path = os.path.join(config["working_dir"], "trainingdata")
        self._data_processor = XMADataProcessor(config)
        self._config = config

    def train(self):
        """Starts training xrommtools-compatible data"""

        mode = self._config["tracking_mode"]
        if mode == "2D":
            try:
                xma_to_dlc(
                    self._config["path_config_file"],
                    self._data_path,
                    self._config["dataset_name"],
                    self._config["experimenter"],
                    self._config["nframes"],
                )
            except UnboundLocalError:
                pass
        elif mode == "per_cam":
            xma_to_dlc(
                path_config_file=self._config["path_config_file"],
                path_config_file_cam2=self._config["path_config_file_2"],
                data_path=self._data_path,
                dataset_name=self._config["dataset_name"],
                scorer=self._config["experimenter"],
                nframes=self._config["nframes"],
                nnetworks=2,
            )
        elif mode == "rgb":
            self._data_processor.make_rgb_video(self._data_path)
        else:
            raise AttributeError(f"Unsupportede mode: {mode}")

        # TODO: instead of having just 1 train function, have separate functions
        # to create training dataset, train model, and evaluate the model.
        deeplabcut.create_training_dataset(self._config["path_config_file"])
        deeplabcut.train_network(
            self._config["path_config_file"], maxiters=self._config["maxiters"]
        )

        if self._config["tracking_mode"] == "per_cam":
            deeplabcut.create_training_dataset(self._config["path_config_file_2"])
            deeplabcut.train_network(
                self._config["path_config_file_2"], maxiters=self._config["maxiters"]
            )



'''Unit tests for XROMM-DLC'''
from pathlib import Path
import shutil
import unittest
from datetime import datetime as dt

import cv2
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from ruamel.yaml import YAML

from deepxromm import DeepXROMM


SAMPLE_FRAME = Path(__file__).parent / 'sample_frame.jpg'
SAMPLE_FRAME_INPUT = Path(__file__).parent / 'sample_frame_input.csv'
SAMPLE_AUTOCORRECT_OUTPUT = Path(__file__).parent / 'sample_autocorrect_output.csv'


class TestProjectCreation(unittest.TestCase):
    '''Tests behaviors related to XMA-DLC project creation'''
    @classmethod
    def setUpClass(cls):
        '''Create a sample project'''
        super(TestProjectCreation, cls).setUpClass()
        cls.project_dir = Path.cwd() / 'tmp'
        DeepXROMM.create_new_project(cls.project_dir)

    def test_project_creates_correct_folders(self):
        '''Do we have all of the correct folders?'''
        for folder in ["trainingdata", "trials"]:
            with self.subTest(folder=folder):
                self.assertTrue((self.project_dir / folder).exists())

    def test_project_creates_config_file(self):
        '''Do we have a project config?'''
        self.assertTrue((self.project_dir / 'project_config.yaml').exists())

    def test_project_config_has_these_variables(self):
        '''Can we access each of the variables that's supposed to be in the config?'''
        yaml = YAML()
        variables = ['task',
        'experimenter',
        'working_dir',
        'path_config_file',
        'dataset_name',
        'nframes',
        'maxiters',
        'tracking_threshold',
        'tracking_mode',
        'swapped_markers',
        'crossed_markers',
        'search_area',
        'threshold',
        'krad',
        'gsigma',
        'img_wt',
        'blur_wt',
        'gamma',
        'cam',
        'frame_num',
        'trial_name',
        'marker',
        'test_autocorrect',
        'cam1s_are_the_same_view']

        yaml = YAML()
        config_path = self.project_dir / 'project_config.yaml'
        with config_path.open() as config:
            project = yaml.load(config)
            for variable in variables:
                with self.subTest(i=variable):
                    self.assertIsNotNone(project[variable])

    @classmethod
    def tearDownClass(cls):
        '''Remove the created temp project'''
        super(TestProjectCreation, cls).tearDownClass()
        shutil.rmtree(cls.project_dir)

class TestDefaultsPerformance(unittest.TestCase):
    '''Test that the config will still be configured properly if the user only provides XMAlab input'''
    def setUp(self):
        '''Create a sample project where the user only inputs XMAlab data'''
        self.working_dir = Path.cwd() / 'tmp'
        self.config = self.working_dir / 'project_config.yaml'
        self.deepxromm = DeepXROMM.create_new_project(self.working_dir)
        frame = cv2.imread(str(SAMPLE_FRAME))

        # Make a trial directory
        (self.working_dir / 'trainingdata/dummy').mkdir(parents=True, exist_ok=True)

        # Cam 1
        video_path_1 = self.working_dir / 'trainingdata/dummy/dummy_cam1.avi'
        out = cv2.VideoWriter(str(video_path_1), cv2.VideoWriter_fourcc(*'DIVX'), 15, (1024, 512))
        out.write(frame)
        out.release()

        # Cam 2
        video_path_2 = self.working_dir / 'trainingdata/dummy/dummy_cam2.avi'
        out = cv2.VideoWriter(str(video_path_2), cv2.VideoWriter_fourcc(*'DIVX'), 15, (1024, 512))
        out.write(frame)
        out.release()

        # CSV
        df = pd.DataFrame({'foo_cam1_X': 0, 'foo_cam1_Y': 0, 'foo_cam2_X': 0, 'foo_cam2_Y': 0,
                        'bar_cam1_X': 0, 'bar_cam1_Y': 0, 'bar_cam2_X': 0, 'bar_cam2_Y': 0,
                        'baz_cam1_X': 0, 'baz_cam1_Y': 0, 'baz_cam2_X': 0, 'baz_cam2_Y': 0}, index=[1])
        csv_path = self.working_dir / 'trainingdata/dummy/dummy.csv'
        df.to_csv(str(csv_path), index=False)
        cv2.destroyAllWindows()

    def test_can_find_frames_from_csv(self):
        '''Can I accurately find the number of frames in the video if the user doesn't tell me?'''
        print(list(self.working_dir.iterdir()))
        deepxromm = DeepXROMM.load_project(self.working_dir)
        config = deepxromm.config
        self.assertEqual(config['nframes'], 1, msg=f"Actual nframes: {config['nframes']}")

    def test_analyze_errors_if_no_folders_in_trials_dir(self):
        '''If there are no trials to analyze, do we return an error?'''
        with self.assertRaises(FileNotFoundError):
            self.deepxromm.analyze_videos()

    def test_autocorrect_errors_if_no_folders_in_trials_dir(self):
        '''If there are no trials to autocorrect, do we return an error?'''
        with self.assertRaises(FileNotFoundError):
            self.deepxromm.autocorrect_trial()

    def test_warn_users_if_nframes_doesnt_match_csv(self):
        '''If the number of frames in the CSV doesn't match the number of frames specified, do I issue a warning?'''
        yaml = YAML()
        with self.config.open() as config:
            tmp = yaml.load(config)

        # Modify the number of frames (similar to how a user would)
        tmp['nframes'] = 2
        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)

        # Check that the user is warned
        with self.assertWarns(UserWarning):
            DeepXROMM.load_project(self.working_dir)

    def test_yaml_file_updates_nframes_after_load_if_frames_is_0(self):
        '''If the user doesn't specify how many frames they want analyzed,
        does their YAML file get updated with how many are in the CSV?'''
        yaml = YAML()
        DeepXROMM.load_project(self.working_dir)
        with self.config.open() as config:
            tmp = yaml.load(config)
        self.assertEqual(tmp['nframes'], 1, msg=f"Actual nframes: {tmp['nframes']}")

    def test_warn_if_user_has_tracked_less_than_threshold_frames(self):
        '''If the user has tracked less than threshold % of their trial,
        do I give them a warning?'''
        DeepXROMM.load_project(self.working_dir)

        # Increase the number of frames in the video to 100 so I can test this
        frame = cv2.imread(str(SAMPLE_FRAME))
        video_path = self.working_dir / 'trainingdata/dummy/dummy_cam1.avi'
        out = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*'DIVX'), 15, (1024, 512))
        for _ in range(100):
            out.write(frame)
        out.release()

        # Check that the user is warned
        with self.assertWarns(UserWarning):
            DeepXROMM.load_project(self.working_dir)

    def test_bodyparts_add_from_csv_if_not_defined(self):
        '''If the user hasn't specified the bodyparts from their trial, we can pull them from the CSV'''
        yaml = YAML()
        date = dt.today().strftime("%Y-%m-%d")
        DeepXROMM.load_project(self.working_dir)
        config_path = Path(self.working_dir) / f'tmp-NA-{date}' / 'config.yaml'

        with config_path.open() as dlc_config:
            config_obj = yaml.load(dlc_config)
        self.assertEqual(config_obj['bodyparts'], ['foo', 'bar', 'baz'])

    def test_bodyparts_add_from_csv_in_3d(self):
        '''If the user wants to do 3D tracking, we output the desired list of bodyparts'''
        yaml = YAML()
        date = dt.today().strftime("%Y-%m-%d")
        with self.config.open('r') as config:
            tmp = yaml.load(config)
        tmp['tracking_mode'] = 'rgb'
        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f'tmp-NA-{date}' / 'config.yaml'
        yaml = YAML()
        with path_to_config.open('r') as dlc_config:
            config_obj = yaml.load(dlc_config)

        self.assertEqual(config_obj['bodyparts'], ['foo_cam1', 'foo_cam2', 'bar_cam1', 'bar_cam2', 'baz_cam1', 'baz_cam2'])

    def test_bodyparts_add_synthetic(self):
        '''Can we add swapped markers?'''
        yaml = YAML()
        date = dt.today().strftime("%Y-%m-%d")

        with self.config.open('r') as config:
            tmp = yaml.load(config)
        tmp['tracking_mode'] = 'rgb'
        tmp['swapped_markers'] = True
        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f'tmp-NA-{date}' / 'config.yaml'
        yaml = YAML()
        with path_to_config.open('r') as dlc_config:
            config_obj = yaml.load(dlc_config)

        self.assertEqual(config_obj['bodyparts'],
                        ['foo_cam1',
                         'foo_cam2',
                         'bar_cam1',
                         'bar_cam2',
                         'baz_cam1',
                         'baz_cam2',
                         'sw_foo_cam1',
                         'sw_foo_cam2',
                         'sw_bar_cam1',
                         'sw_bar_cam2',
                         'sw_baz_cam1',
                         'sw_baz_cam2'])

    def test_bodyparts_add_crossed(self):
        '''Can we add crossed markers?'''
        yaml = YAML()
        date = dt.today().strftime("%Y-%m-%d")

        with self.config.open('r') as config:
            tmp = yaml.load(config)
        tmp['tracking_mode'] = 'rgb'
        tmp['crossed_markers'] = True
        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f'tmp-NA-{date}' / 'config.yaml'
        yaml = YAML()
        with path_to_config.open('r') as dlc_config:
            config_obj = yaml.load(dlc_config)

        self.assertEqual(config_obj['bodyparts'], ['foo_cam1', 'foo_cam2', 'bar_cam1', 'bar_cam2', 'baz_cam1', 'baz_cam2', 'cx_foo_cam1x2', 'cx_bar_cam1x2', 'cx_baz_cam1x2'])

    def test_bodyparts_add_synthetic_and_crossed(self):
        '''Can we add both swapped and crossed markers?'''
        yaml = YAML()
        date = dt.today().strftime("%Y-%m-%d")

        with self.config.open('r') as config:
            tmp = yaml.load(config)
        tmp['tracking_mode'] = 'rgb'
        tmp['swapped_markers'] = True
        tmp['crossed_markers'] = True
        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f'tmp-NA-{date}' / 'config.yaml'
        yaml = YAML()
        with path_to_config.open('r') as dlc_config:
            config_obj = yaml.load(dlc_config)

        self.assertEqual(config_obj['bodyparts'],
                        ['foo_cam1',
                         'foo_cam2',
                         'bar_cam1',
                         'bar_cam2',
                         'baz_cam1',
                         'baz_cam2',
                         'sw_foo_cam1',
                         'sw_foo_cam2',
                         'sw_bar_cam1',
                         'sw_bar_cam2',
                         'sw_baz_cam1',
                         'sw_baz_cam2',
                         'cx_foo_cam1x2',
                         'cx_bar_cam1x2',
                         'cx_baz_cam1x2'])

    def test_bodyparts_error_if_different_from_csv(self):
        '''If the user specifies different bodyparts than their CSV, raise an error'''
        yaml = YAML()
        date = dt.today().strftime("%Y-%m-%d")
        path_to_config = self.working_dir / f'tmp-NA-{date}' / 'config.yaml'
        yaml = YAML()
        with path_to_config.open('r') as dlc_config:
            config_dlc = yaml.load(dlc_config)

        config_dlc['bodyparts'] = ['foo', 'bar']

        with path_to_config.open('w') as dlc_config:
            yaml.dump(config_dlc, dlc_config)

        with self.assertRaises(SyntaxError):
            DeepXROMM.load_project(self.working_dir)

    def test_autocorrect_error_if_trial_not_set(self):
        '''If the user doesn't specify a trial to test autocorrect with, do we error?'''
        yaml = YAML()
        with self.config.open('r') as config:
            tmp = yaml.load(config)

        tmp['test_autocorrect'] = True
        tmp['marker'] = 'foo'
        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)

        with self.assertRaises(SyntaxError):
            DeepXROMM.load_project(self.working_dir)

    def test_autocorrect_error_if_marker_not_set(self):
        '''If the user doesn't specify a marker to test autocorrect with, do we error?'''
        yaml = YAML()
        
        with self.config.open('r') as config:
            tmp = yaml.load(config)

        tmp['test_autocorrect'] = True
        tmp['trial_name'] = 'test'

        with self.config.open('w') as fp:
            yaml.dump(tmp, fp)

        with self.assertRaises(SyntaxError):
            DeepXROMM.load_project(self.working_dir)

    def tearDown(self):
        '''Remove the created temp project'''
        shutil.rmtree(self.working_dir)

class TestSampleTrial(unittest.TestCase):
    '''Test function behaviors using a frame from an actual trial'''
    def setUp(self):
        '''Create trial'''
        self.working_dir = Path.cwd() / 'tmp'
        self.deepxromm = DeepXROMM.create_new_project(self.working_dir)
        frame = cv2.imread(str(SAMPLE_FRAME))

        # Make a trial directory
        trial_dir = self.working_dir / 'trainingdata/test'
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Move sample frame input to trainingdata
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(trial_dir / 'test.csv'))

        # Cam 1 and Cam 2 trainingdata setup
        for cam in ['cam1', 'cam2']:
            video_path = trial_dir / f'dummy_{cam}.avi'
            out = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*'DIVX'), 30, (1024, 512))
            out.write(frame)
            out.release()

        # Move sample frame input to trials (it0 and trials)
        trials_dir = self.working_dir / 'trials/test/it0'
        trials_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(self.working_dir / 'trials/test/test.csv'))
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(trials_dir / 'test-Predicted2DPoints.csv'))

        # Cam 1 and Cam 2 trials setup
        for cam in ['cam1', 'cam2']:
            video_path = self.working_dir / f'trials/test/test_{cam}.avi'
            out = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*'DIVX'), 30, (1024, 512))
            out.write(frame)
            out.release()

        cv2.destroyAllWindows()

    def test_autocorrect_is_working(self):
        '''Make sure that autocorrect still works properly after making changes'''
        # Run autocorrect on the sample frame
        self.deepxromm.autocorrect_trial()

        # Load CSVs
        function_output_path = self.working_dir / 'trials/test/it0/test-AutoCorrected2DPoints.csv'
        function_output = pd.read_csv(str(function_output_path), dtype='float64')
        sample_output = pd.read_csv(str(SAMPLE_AUTOCORRECT_OUTPUT), dtype='float64')

        # Drop cam2 markers and check for changes
        columns_to_drop = function_output.columns[function_output.columns.str.contains('cam2', case=False)]
        function_output.drop(columns_to_drop, axis=1, inplace=True)
        function_output = function_output.round(6)
        sample_output = sample_output.round(6)

        try:
            # Use assert_frame_equal to check if the data frames are the same
            assert_frame_equal(function_output, sample_output, check_exact=False, rtol=1e-6, atol=1e-6)
        except AssertionError as e:
            print(f"Autocorrector diff: {function_output.compare(sample_output)}")
            raise e

    def test_image_hashing_identical_trials_returns_0(self):
        '''Make sure the image hashing function is working properly'''
        # Create an identical second trial
        frame_path = self.working_dir.parent / 'sample_frame.jpg'
        frame = cv2.imread(str(frame_path))
        trial_dir = self.working_dir / 'trials/test2_same'
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Create videos
        for cam in ['cam1', 'cam2']:
            video_path = trial_dir / f'test2_same_{cam}.avi'
            out = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*'DIVX'), 30, (1024, 512))
            out.write(frame)
            out.release()

        # Do similarity comparison
        similarity = self.deepxromm.analyze_video_similarity_project()
        # Since both videos are the same, the image similarity output should be 0
        self.assertEqual(sum(similarity.values()), 0)

    def test_image_hashing_different_trials_returns_nonzero(self):
        '''Image hashing different videos returns nonzero answer'''
        # Create a different second trial
        frame = np.zeros((480, 480, 3), np.uint8)
        trial_dir = self.working_dir / 'trials/test2_diff'
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Create videos
        for cam in ['cam1', 'cam2']:
            video_path = trial_dir / f'test2_diff_{cam}.avi'
            out = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*'DIVX'), 30, (480, 480))
            out.write(frame)
            out.release()

        # Do similarity comparison
        similarity = self.deepxromm.analyze_video_similarity_project()
        # Since the videos are different, should return nonzero answer
        self.assertNotEqual(sum(similarity.values()), 0)

    def test_marker_similarity_returns_0_if_identical(self):
        '''Check that identical data has a similarity value of 0'''
        # Move sample data into test trial
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(self.working_dir / 'trials/test/test.csv'))

        # Move sample data into test2 trial
        (self.working_dir / 'trials/test2').mkdir(parents=True, exist_ok=True)
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(self.working_dir / 'trials/test2/test2.csv'))

        # Do cross-correlation
        marker_similarity = self.deepxromm.analyze_marker_similarity_project()
        self.assertEqual(sum(marker_similarity.values()), 0)

    def test_marker_similarity_returns_not_0_if_different(self):
        '''Check that different data has a similarity value of not 0'''
        # Move sample data into test trial
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(self.working_dir / 'trials/test/test.csv'))

        # Move autocorrect data into test2 trial
        (self.working_dir / 'trials/test2').mkdir(parents=True, exist_ok=True)
        shutil.copy(str(SAMPLE_AUTOCORRECT_OUTPUT), str(self.working_dir / 'trials/test2/test2.csv'))

        # Do cross-correlation
        marker_similarity = self.deepxromm.analyze_marker_similarity_project()
        self.assertNotEqual(sum(marker_similarity.values()), 0)

    def tearDown(self):
        '''Remove the created temp project'''
        project_path = Path.cwd() / 'tmp'
        shutil.rmtree(project_path)

if __name__ == "__main__":
    unittest.main()



"""Converts data from XMALab into the format useful for DLC training."""

import glob
import os

from subprocess import Popen

import blend_modes
import cv2
from PIL import Image
import numpy as np
import pandas as pd
from ruamel.yaml import YAML


class XMADataProcessor:
    """Converts data from XMALab into the format useful for DLC training."""

    def __init__(self, config):
        self._config = config
        self._swap_markers = config["swapped_markers"]
        self._cross_markers = config["crossed_markers"]

    def get_bodyparts_from_xma(self, path_to_trial, mode):
        """Pulls the names of the XMAlab markers from the 2Dpoints file"""

        csv_path = [file for file in os.listdir(path_to_trial) if file[-4:] == ".csv"]
        if len(csv_path) > 1:
            raise FileExistsError(
                "Found more than 1 CSV file for trial: " + path_to_trial
            )
        if len(csv_path) <= 0:
            raise FileNotFoundError(
                "Couldn't find a CSV file for trial: " + path_to_trial
            )
        trial_csv = pd.read_csv(
            os.path.join(path_to_trial, csv_path[0]),
            sep=",",
            header=0,
            dtype="float",
            na_values="NaN",
        )
        names = trial_csv.columns.values
        if mode == "rgb":
            parts = [name.rsplit("_", 1)[0] for name in names]
            if self._swap_markers:
                parts = parts + [f"sw_{part}" for part in parts]
            if self._cross_markers:
                parts = parts + [
                    f"cx_{part}_cam1x2"
                    for part in [name.rsplit("_", 2)[0] for name in names]
                ]
        elif mode in ["2D", "per_cam"]:
            parts = [name.rsplit("_", 2)[0] for name in names]
        else:
            raise SyntaxError("Invalid value for mode parameter")

        parts_unique = []
        for part in parts:
            if part not in parts_unique:
                parts_unique.append(part)
        return parts_unique

    def make_rgb_video(self, data_path):
        """For all trials in given data path merges 2 videos into single RBG video."""
        trials = [
            folder
            for folder in os.listdir(data_path)
            if os.path.isdir(os.path.join(data_path, folder))
            and not folder.startswith(".")
        ]
        for trial in trials:
            path_to_trial = f"{data_path}/{trial}"
            self._merge_rgb(path_to_trial)
            substitute_data_relpath = "labeled-data/" + self._config["dataset_name"]
            substitute_data_abspath = os.path.join(
                os.path.split(self._config["path_config_file"])[0],
                substitute_data_relpath,
            )
            # TODO: decouple the behavior for videos vs data points/image
            # extraction. extract_matched_frames works for trials with an
            # XMAlab-formatted CSV attached (typically training data), while merge_rgb
            # works on all videos as long as there's a cam 1 and a cam2 video.
            self._extract_matched_frames_rgb(
                path_to_trial,
                substitute_data_abspath,
                range(1, self._config["nframes"] + 1),
            )
            self._splice_xma_to_dlc(path_to_trial)

    def find_cam_file(self, path, identifier):
        """Searches a file for a given cam video in the trail folder."""
        files = glob.glob(os.path.join(path, f"*{identifier}*.avi"))
        if files:
            result = files[0]
            print(f"Found file {result} for {identifier}")
            return files[0]
        else:
            raise FileNotFoundError(f"No video file found containing '{identifier}' in {path}")

    def list_trials(self):
        path = os.path.join(self._config["working_dir"], "trials")
        """Returns a list of trials or throws an exception if folder is empty"""
        trials = [folder for folder in os.listdir(path) if os.path.isdir(os.path.join(path, folder)) and not folder.startswith('.')]

        if len(trials) <= 0:
            raise FileNotFoundError(f'Empty trials directory found. Please put trials to be '
                                     'analyzed after training into the {path} folder')
        return trials

    def split_rgb(self, trial_path, codec='avc1'):
        '''Takes a RGB video with different grayscale data written to the R, G, and B channels and splits it back into its component source videos.'''
        trial_name = os.path.basename(os.path.normpath(trial_path))
        out_name = trial_name + '_split_'

        try:
            rgb_video = cv2.VideoCapture(f'{trial_path}/{trial_name}_rgb.avi')
        except FileNotFoundError as e:
            print(f'Couldn\'t find video at {trial_path}/{trial_name}_rgb.avi')
            raise e
        frame_width = int(rgb_video.get(3))
        frame_height = int(rgb_video.get(4))
        frame_rate = round(rgb_video.get(5),2)
        if codec == 'uncompressed':
            pix_format = 'gray'   ##change to 'yuv420p' for color or 'gray' for grayscale. 'pal8' doesn't play on macs
            cam1_split_ffmpeg = Popen(['ffmpeg', '-y', '-f', 'image2pipe', '-vcodec', 'png', '-r', str(int(frame_rate)),
            '-i', '-', '-vcodec', 'rawvideo','-pix_fmt',pix_format,'-r', str(int(frame_rate)), f'{trial_path}/{out_name}'+'cam1.avi'], stdin=PIPE)
            cam2_split_ffmpeg = Popen(['ffmpeg', '-y', '-f', 'image2pipe', '-vcodec', 'png', '-r', str(int(frame_rate)),
            '-i', '-', '-vcodec', 'rawvideo','-pix_fmt',pix_format,'-r', str(int(frame_rate)), f'{trial_path}/{out_name}'+'cam2.avi'], stdin=PIPE)
            blue_split_ffmpeg = Popen(['ffmpeg', '-y', '-f', 'image2pipe', '-vcodec', 'png', '-r', str(int(frame_rate)),
            '-i', '-', '-vcodec', 'rawvideo','-pix_fmt',pix_format,'-r', str(int(frame_rate)), f'{trial_path}/{out_name}'+'blue.avi'], stdin=PIPE)
        else:
            if codec == 0:
                fourcc = 0
            else:
                fourcc = cv2.VideoWriter_fourcc(*codec)
            cam1 = cv2.VideoWriter(f'{trial_path}/{out_name}'+'cam1.avi',
                                    fourcc,
                                    frame_rate,(frame_width, frame_height))
            cam2 = cv2.VideoWriter(f'{trial_path}/{out_name}'+'cam2.avi',
                                    fourcc,
                                    frame_rate,(frame_width, frame_height))
            blue_channel = cv2.VideoWriter(f'{trial_path}/{out_name}'+'blue.avi',
                                    fourcc,
                                    frame_rate,(frame_width, frame_height))

        i = 1
        while rgb_video.isOpened():
            ret, frame = rgb_video.read()
            print(f'Current Frame: {i}')
            i = i + 1
            if ret:
                B, G, R = cv2.split(frame)
                if codec == 'uncompressed':
                    im_r = Image.fromarray(R)
                    im_g = Image.fromarray(G)
                    im_b = Image.fromarray(B)
                    im_r.save(cam1_split_ffmpeg.stdin, 'PNG')
                    im_g.save(cam2_split_ffmpeg.stdin, 'PNG')
                    im_b.save(blue_split_ffmpeg.stdin, 'PNG')
                    if cv2.waitKey(1) & 0xFF == ord('q'):
                        break
                else:
                    cam1.write(R)
                    cam2.write(G)
                    blue_channel.write(B)
                    if cv2.waitKey(1) & 0xFF == ord('q'):
                        break
            else:
                break
        if codec == 'uncompressed':
            cam1_split_ffmpeg.stdin.close()
            cam1_split_ffmpeg.wait()
            cam2_split_ffmpeg.stdin.close()
            cam2_split_ffmpeg.wait()
            blue_split_ffmpeg.stdin.close()
            blue_split_ffmpeg.wait()
        else:
            cam1.release()
            cam2.release()
            blue_channel.release()
        rgb_video.release()
        cv2.destroyAllWindows()
        print(f"Cam1 grayscale video created at {trial_path}/{out_name}cam1.avi!")
        print(f"Cam2 grayscale video created at {trial_path}/{out_name}cam2.avi!")
        print(f"Blue channel grayscale video created at {trial_path}/{out_name}blue.avi!")

    def _merge_rgb(self, trial_path, codec="avc1", mode="difference"):
        """
        Takes the path to a trial subfolder and exports a single new video with
        cam1 video written to the red channel and cam2 video written to the
        green channel. The blue channel is, depending on the value of config
        "mode", either the difference blend between A and B, the multiply
        blend, or just a black frame.
        """
        trial_name = os.path.basename(os.path.normpath(trial_path))
        if os.path.exists(f"{trial_path}/{trial_name}_rgb.avi"):
            print("RGB video already created. Skipping.")
            return
        cam1_video_path = self.find_cam_file(trial_path, "cam1")
        cam1_video = cv2.VideoCapture(cam1_video_path)

        cam2_video_path = self.find_cam_file(trial_path, "cam2")
        cam2_video = cv2.VideoCapture(cam2_video_path)

        frame_width = int(cam1_video.get(3))
        frame_height = int(cam1_video.get(4))
        frame_rate = round(cam1_video.get(5), 2)
        fourcc = cv2.VideoWriter_fourcc(*codec)
        out = cv2.VideoWriter(
            f"{trial_path}/{trial_name}_rgb.avi",
            fourcc,
            frame_rate,
            (frame_width, frame_height),
        )
        i = 1
        while cam1_video.isOpened():
            print(f"Current Frame: {i}")
            ret_cam1, frame_cam1 = cam1_video.read()
            _, frame_cam2 = cam2_video.read()
            if ret_cam1:
                frame_cam1 = cv2.cvtColor(frame_cam1, cv2.COLOR_BGR2BGRA, 4).astype(
                    np.float32
                )
                frame_cam2 = cv2.cvtColor(frame_cam2, cv2.COLOR_BGR2BGRA, 4).astype(
                    np.float32
                )
                frame_cam1 = cv2.normalize(
                    frame_cam1, None, 0, 255, norm_type=cv2.NORM_MINMAX
                )
                frame_cam2 = cv2.normalize(
                    frame_cam2, None, 0, 255, norm_type=cv2.NORM_MINMAX
                )
                if mode == "difference":
                    extra_channel = blend_modes.difference(frame_cam1, frame_cam2, 1)
                elif mode == "multiply":
                    extra_channel = blend_modes.multiply(frame_cam1, frame_cam2, 1)
                else:
                    extra_channel = np.zeros((frame_width, frame_height, 3), np.uint8)
                    extra_channel = cv2.cvtColor(
                        extra_channel, cv2.COLOR_BGR2BGRA, 4
                    ).astype(np.float32)
                frame_cam1 = cv2.cvtColor(frame_cam1, cv2.COLOR_BGRA2BGR).astype(
                    np.uint8
                )
                frame_cam2 = cv2.cvtColor(frame_cam2, cv2.COLOR_BGRA2BGR).astype(
                    np.uint8
                )
                extra_channel = cv2.cvtColor(extra_channel, cv2.COLOR_BGRA2BGR).astype(
                    np.uint8
                )
                frame_cam1 = cv2.cvtColor(frame_cam1, cv2.COLOR_BGR2GRAY)
                frame_cam2 = cv2.cvtColor(frame_cam2, cv2.COLOR_BGR2GRAY)
                extra_channel = cv2.cvtColor(extra_channel, cv2.COLOR_BGR2GRAY)
                merged = cv2.merge((extra_channel, frame_cam2, frame_cam1))
                out.write(merged)
                if cv2.waitKey(1) & 0xFF == ord("q"):
                    break
            else:
                break

            i = i + 1
        cam1_video.release()
        cam2_video.release()
        out.release()
        cv2.destroyAllWindows()
        print(f"Merged RGB video created at {trial_path}/{trial_name}_rgb.avi!")

    def _splice_xma_to_dlc(self, trial_path, outlier_mode=False):
        """Takes csv of XMALab 2D XY coordinates from 2 cameras, outputs spliced hdf+csv data for DeepLabCut"""
        substitute_data_relpath = "labeled-data/" + self._config["dataset_name"]
        substitute_data_abspath = os.path.join(
            os.path.sep.join(self._config["path_config_file"].split("\\")[:-1]),
            substitute_data_relpath,
        )
        markers = self.get_bodyparts_from_xma(trial_path, mode='2D')
        # TODO: this entire section can be solved with a creative call to
        # get_bodyparts_from_xma and some dataFrame manipulation to be
        # significantly shorter (and potentially faster?)
        try:
            trial_name = os.path.basename(os.path.normpath(trial_path))
            df = pd.read_csv(f"{trial_path}/{trial_name}.csv")
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Please make sure that your trainingdata 2DPoints csv file is named {trial_name}.csv"
            ) from e
        if self._swap_markers:
            print("Creating cam1Y-cam2Y-swapped synthetic markers")
            swaps = []
            df_sw = pd.DataFrame()
            for marker in markers:
                name_x1 = marker + "_cam1_X"
                name_x2 = marker + "_cam2_X"
                name_y1 = marker + "_cam1_Y"
                name_y2 = marker + "_cam2_Y"
                swap_name_x1 = "sw_" + name_x1
                swap_name_x2 = "sw_" + name_x2
                swap_name_y1 = "sw_" + name_y1
                swap_name_y2 = "sw_" + name_y2
                df_sw[swap_name_x1] = df[name_x1]
                df_sw[swap_name_y1] = df[name_y2]
                df_sw[swap_name_x2] = df[name_x2]
                df_sw[swap_name_y2] = df[name_y1]
                swaps.extend([swap_name_x1, swap_name_y1, swap_name_x2, swap_name_y2])
            df = df.join(df_sw)
            print(swaps)
        if self._cross_markers:
            print("Creating cam1-cam2-crossed synthetic markers")
            crosses = []
            df_cx = pd.DataFrame()
            for marker in markers:
                name_x1 = marker + "_cam1_X"
                name_x2 = marker + "_cam2_X"
                name_y1 = marker + "_cam1_Y"
                name_y2 = marker + "_cam2_Y"
                cross_name_x = "cx_" + marker + "_cam1x2_X"
                cross_name_y = "cx_" + marker + "_cam1x2_Y"
                df_cx[cross_name_x] = df[name_x1] * df[name_x2]
                df_cx[cross_name_y] = df[name_y1] * df[name_y2]
                crosses.extend([cross_name_x, cross_name_y])
            df = df.join(df_cx)
            print(crosses)
        names_final = df.columns.values
        parts_final = [name.rsplit("_", 1)[0] for name in names_final]
        parts_unique_final = []
        for part in parts_final:
            if not part in parts_unique_final:
                parts_unique_final.append(part)
        print("Importing markers: ")
        print(parts_unique_final)
        with open(self._config["path_config_file"], "r") as dlc_config:
            yaml = YAML()
            dlc_proj = yaml.load(dlc_config)

        dlc_proj["bodyparts"] = parts_unique_final

        with open(self._config["path_config_file"], "w") as dlc_config:
            yaml.dump(dlc_proj, dlc_config)

        df = df.dropna(how="all")
        unique_frames_set = {}
        unique_frames_set = {
            index
            for index in range(1, self._config["nframes"] + 1)
            if index not in unique_frames_set
        }
        unique_frames = sorted(unique_frames_set)
        print("Importing frames: ")
        print(unique_frames)
        df["frame_index"] = [
            substitute_data_relpath
            + f"/{trial_name}_rgb_"
            + str(index).zfill(4)
            + ".png"
            for index in unique_frames
        ]
        df["scorer"] = self._config["experimenter"]
        df = df.melt(id_vars=["frame_index", "scorer"])
        new = df["variable"].str.rsplit("_", n=1, expand=True)
        df["variable"], df["coords"] = new[0], new[1]
        df = df.rename(columns={"variable": "bodyparts"})
        df["coords"] = df["coords"].str.rstrip(" ").str.lower()
        cat_type = pd.api.types.CategoricalDtype(
            categories=parts_unique_final, ordered=True
        )
        df["bodyparts"] = df["bodyparts"].str.lstrip(" ").astype(cat_type)
        newdf = df.pivot_table(
            columns=["scorer", "bodyparts", "coords"],
            index="frame_index",
            values="value",
            aggfunc="first",
            dropna=False,
        )
        newdf.index.name = None
        if not os.path.exists(substitute_data_abspath):
            os.makedirs(substitute_data_abspath)
        if outlier_mode:
            data_name = os.path.join(substitute_data_abspath, "MachineLabelsRefine.h5")
            tracked_hdf = os.path.join(
                substitute_data_abspath, ("MachineLabelsRefine_" + ".h5")
            )
        else:
            data_name = os.path.join(
                substitute_data_abspath,
                ("CollectedData_" + self._config["experimenter"] + ".h5"),
            )
            tracked_hdf = os.path.join(
                substitute_data_abspath,
                ("CollectedData_" + self._config["experimenter"] + ".h5"),
            )
        newdf.to_hdf(data_name, "df_with_missing", format="table", mode="w")
        newdf.to_hdf(tracked_hdf, "df_with_missing", format="table", mode="w")
        tracked_csv = data_name.split(".h5")[0] + ".csv"
        newdf.to_csv(tracked_csv, na_rep="NaN")
        print(
            "Successfully spliced XMALab 2D points to DLC format",
            "saved " + str(data_name),
            "saved " + str(tracked_hdf),
            "saved " + str(tracked_csv),
            sep="\n",
        )

    def _extract_matched_frames_rgb(
        self, trial_path, labeled_data_path, indices, compression=1
    ):
        """Given a list of frame indices and a project path, produce a folder (in labeled-data) of matching frame pngs per source video.
        Optionally, compress the output PNGs. Factor ranges from 0 (no compression) to 9 (most compression)"""
        extracted_frames = []
        trainingdata_path = self._config["working_dir"] + "/trainingdata"
        trial_name = os.path.basename(os.path.normpath(trial_path))
        video_path = f"{trainingdata_path}/{trial_name}/{trial_name}_rgb.avi"
        labeled_data_path = (
            os.path.split(self._config["path_config_file"])[0]
            + "/labeled-data/"
            + self._config["dataset_name"]
        )
        frames_from_vid = self._vid_to_pngs(
            video_path,
            indices,
            labeled_data_path,
            name_from_folder=True,
            compression=compression,
        )
        extracted_frames.append(frames_from_vid)
        print("Extracted " + str(len(indices)) + f" matching frames from {video_path}")

    def _vid_to_pngs(
        self,
        video_path,
        indices_to_match,
        output_dir=None,
        name_from_folder=True,
        compression=0,
    ):
        """Takes a list of frame numbers and exports matching frames from a video as pngs.
        Optionally, compress the output PNGs. Factor ranges from 0 (no compression) to 9 (most compression)"""
        frame_index = 1
        last_frame_to_analyze = max(indices_to_match)
        png_list = []
        if name_from_folder:
            out_name = os.path.splitext(os.path.basename(video_path))[0]
        else:
            out_name = "img"
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        print("===== try this")
        cap = cv2.VideoCapture(video_path)
        print("===== end this")
        while cap.isOpened():
            ret, frame = cap.read()
            if ret is False:
                break
            if frame_index > last_frame_to_analyze:
                break
            if frame_index not in indices_to_match:
                frame_index += 1
                continue

            print(f"Extracting frame {frame_index}")
            png_name = out_name + "_" + str(frame_index).zfill(4) + ".png"
            png_path = os.path.join(output_dir, png_name)
            png_list.append(png_path)
            cv2.imwrite(png_path, frame, [cv2.IMWRITE_PNG_COMPRESSION, compression])
            if cv2.waitKey(1) & 0xFF == ord("q"):
                break
            frame_index += 1

        cap.release()
        cv2.destroyAllWindows()
        return png_list



'''Taken from https://stackoverflow.com/questions/5411603/how-to-remove-trailing-whitespace-in-code-using-another-script'''

import os

PATH = os.getcwd()

for path, dirs, files in os.walk(PATH):
    for f in files:
        file_name, file_extension = os.path.splitext(f)
        if file_extension == '.py':
            path_name = os.path.join(path, f)
            with open(path_name, 'r') as fh:
                new = [line.rstrip() for line in fh]
            with open(path_name, 'w') as fh:
                [fh.write('%s\n' % line) for line in new]
